class DummyPanel:
    """A dummy lab panel object."""

    def __init__(self, name=None, verbose_name=None, requisition_model=None):
        self.name = name
        self.verbose_name = verbose_name or name
        self.requisition_model = requisition_model

    def __str__(self):
        return self.verbose_name

    def __repr__(self):
        return f"{self.__class__.__name__}(name='{self.name}')"



from __future__ import annotations

import re
from decimal import Decimal
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from edc_facility.utils import get_default_facility_name, get_facility
from edc_utils import get_utcnow, to_utc

from .crf_collection import CrfCollection
from .forms_collection import FormsCollection
from .requisition_collection import RequisitionCollection
from .window_period import WindowPeriod

if TYPE_CHECKING:
    from datetime import datetime

    from dateutil.relativedelta import relativedelta
    from edc_facility import Facility

    from .crf import Crf
    from .requisition import Requisition


class VisitCodeError(Exception):
    pass


class VisitDateError(Exception):
    pass


class VisitError(Exception):
    pass


class BaseDatetimeNotSet(Exception):
    pass


def base_datetime_required():
    raise BaseDatetimeNotSet(
        "Base datetime is None, set the base datetime before accessing attr"
    )


class VisitDate:
    window_period_cls = WindowPeriod

    def __init__(
        self,
        rlower: relativedelta = None,
        rupper: relativedelta = None,
        timepoint: Decimal = None,
        base_timepoint: Decimal = None,
    ):
        self._base: datetime | None = None
        self._lower: datetime | None = None
        self._upper: datetime | None = None
        self._window_period = self.window_period_cls(
            rlower=rlower,
            rupper=rupper,
            timepoint=timepoint,
            base_timepoint=base_timepoint,
        )

    @property
    def base(self) -> datetime:
        return self._base

    @base.setter
    def base(self, dt: datetime = None):
        self._base = to_utc(dt)
        self._lower, self._upper = self._window_period.get_window(dt=self._base)

    @property
    def lower(self) -> datetime:
        if not self.base:
            raise BaseDatetimeNotSet(
                "Base datetime is None, set the base datetime before accessing attr lower"
            )
        return self._lower

    @property
    def upper(self) -> datetime:
        if not self.base:
            raise BaseDatetimeNotSet(
                "Base datetime is None, set the base datetime before accessing attr upper"
            )
        return self._upper


class Visit:
    code_regex = r"^([A-Z0-9])+$"
    visit_date_cls = VisitDate

    def __init__(
        self,
        code: str = None,
        timepoint: int | float | Decimal = None,
        rbase: relativedelta = None,
        rlower: relativedelta = None,
        rupper: relativedelta = None,
        rlower_late: relativedelta = None,
        rupper_late: relativedelta = None,
        add_window_gap_to_lower: bool | None = None,
        max_window_gap_to_lower: int | None = None,
        crfs: CrfCollection | None = None,
        requisitions: RequisitionCollection | None = None,
        crfs_unscheduled: CrfCollection | None = None,
        crfs_missed: CrfCollection | None = None,
        requisitions_unscheduled: RequisitionCollection | None = None,
        crfs_prn: CrfCollection | None = None,
        requisitions_prn: RequisitionCollection | None = None,
        title: str = None,
        allow_unscheduled: bool | None = None,
        facility_name: str | None = None,
        instructions: str | None = None,
        base_timepoint: int | float | Decimal | None = None,
        grouping=None,
    ):
        self.next = None
        if isinstance(base_timepoint, (float,)):
            base_timepoint = Decimal(str(base_timepoint))
        elif isinstance(base_timepoint, (int,)):
            base_timepoint = Decimal(str(base_timepoint) + ".0")
        self.base_timepoint = base_timepoint or Decimal("0.0")
        self.crfs: CrfCollection = crfs or CrfCollection()
        self.crfs_unscheduled: CrfCollection = crfs_unscheduled or CrfCollection()
        self.crfs_missed: CrfCollection = crfs_missed or CrfCollection()
        self.crfs_prn: CrfCollection = crfs_prn or CrfCollection()
        for prn in self.crfs_prn:
            prn.required = False
        self.requisitions: RequisitionCollection = requisitions or RequisitionCollection()
        self.requisitions_unscheduled: RequisitionCollection = (
            requisitions_unscheduled or RequisitionCollection()
        )
        self.requisitions_prn: RequisitionCollection = (
            requisitions_prn or RequisitionCollection()
        )
        for prn in self.requisitions_prn:
            prn.required = False
        self.instructions = instructions
        if isinstance(timepoint, (float,)):
            timepoint = Decimal(str(timepoint))
        elif isinstance(timepoint, (int,)):
            timepoint = Decimal(str(timepoint) + ".0")
        self.timepoint = timepoint
        self.rbase = rbase
        self.rlower = rlower
        self.rupper = rupper
        self.rlower_late = self.rlower if rlower_late is None else rlower_late
        self.rupper_late = self.rupper if rupper_late is None else rupper_late
        self.add_window_gap_to_lower = add_window_gap_to_lower
        self.max_window_gap_to_lower = max_window_gap_to_lower
        self.grouping = grouping
        if not code or isinstance(code, int) or not re.match(self.code_regex, code):
            raise VisitCodeError(f"Invalid visit code. Got '{code}'")
        else:
            self.code = code  # unique
        self.dates = self.visit_date_cls(
            rlower=rlower,
            rupper=rupper,
            timepoint=self.timepoint,
            base_timepoint=self.base_timepoint,
        )
        self.late_dates = self.visit_date_cls(
            rlower=rlower_late or rlower,
            rupper=rupper_late or rupper,
            timepoint=self.timepoint,
            base_timepoint=self.base_timepoint,
        )
        utcnow = get_utcnow()
        self.validate_rlower_late(utcnow)
        self.validate_rupper_late(utcnow)
        self.title = title or f"Visit {self.code}"
        self.name = self.code
        self.facility_name = facility_name or get_default_facility_name()
        self.allow_unscheduled = allow_unscheduled
        if timepoint is None:
            raise VisitError(f"Timepoint not specified. Got None. See Visit {code}.")

    def __repr__(self):
        return f"{self.__class__.__name__}({self.code}, {self.timepoint})"

    def __str__(self):
        return self.title

    @property
    def scheduled_forms(self) -> FormsCollection:
        """Returns a FormsCollection of scheduled forms."""
        return FormsCollection(*self.crfs, *self.requisitions, name="scheduled_forms")

    @property
    def unscheduled_forms(self) -> FormsCollection:
        """Returns a FormsCollection of unscheduled forms."""
        return FormsCollection(
            *self.crfs_unscheduled, *self.requisitions_unscheduled, name="unscheduled_forms"
        )

    @property
    def prn_forms(self) -> FormsCollection:
        """Returns a FormsCollection of prn forms."""
        return FormsCollection(*self.crfs_prn, *self.requisitions_prn, name="prn_forms")

    @property
    def all_crfs(self) -> CrfCollection:
        crfs = list(self.crfs) + [
            crf
            for crf in self.crfs_unscheduled
            if crf.model not in [crf.model for crf in self.crfs]
        ]
        crfs = crfs + [
            crf for crf in self.crfs_missed if crf.model not in [crf.model for crf in crfs]
        ]
        crfs = crfs + [
            crf for crf in self.crfs_prn if crf.model not in [crf.model for crf in crfs]
        ]
        return CrfCollection(*crfs, name="all_crfs", check_sequence=False)

    @property
    def all_requisitions(self) -> RequisitionCollection:
        names = [r.name for r in self.requisitions]
        requisitions = list(self.requisitions) + [
            r for r in self.requisitions_unscheduled if r.name not in names
        ]
        names = list(set([r.name for r in requisitions]))
        requisitions = requisitions + [r for r in self.requisitions_prn if r.name not in names]
        return RequisitionCollection(
            *requisitions, name="all_requisitions", check_sequence=False
        )

    # def next_form(
    #     self, model: str = None, panel: str | None = None
    # ) -> Crf | Requisition | None:
    #     """Returns the next required and scheduled "form" or None."""
    #     next_form = None
    #     for index, form in enumerate(self.scheduled_forms):
    #         if form.model == model and form.required and getattr(form, "panel", "") == panel:
    #             try:
    #                 next_form = self.scheduled_forms.forms[index + 1]
    #             except IndexError:
    #                 pass
    #         elif form.model == model and form.required:
    #             try:
    #                 next_form = self.scheduled_forms.forms[index + 1]
    #             except IndexError:
    #                 pass
    #     return next_form

    # def get_form(self, model=None) -> Crf | Requisition | None:
    #     """Returns a schedule form or none"""
    #     for form in self.scheduled_forms:
    #         if form.model == model:
    #             return form
    #     return None

    def get_crf(self, model=None) -> Crf | None:
        get_crf = None
        for crf in self.crfs:
            if crf.model == model:
                get_crf = crf
                break
        return get_crf

    def get_requisition(self, model=None, panel_name=None) -> Requisition | None:
        get_requisition = None
        for requisition in self.requisitions:
            if requisition.model == model and requisition.panel.name == panel_name:
                get_requisition = requisition
                break
        return get_requisition

    def get_models(self) -> list:
        models = []
        for crf in self.crfs:
            models.append(django_apps.get_model(crf.model))
        for crf in self.requisitions:
            models.append(django_apps.get_model(crf.model))
        return models

    @property
    def facility(self) -> Facility | None:
        """Returns a Facility object."""
        if self.facility_name:
            return get_facility(name=self.facility_name)
        return None

    @property
    def timepoint_datetime(self) -> datetime:
        return self.dates.base

    @timepoint_datetime.setter
    def timepoint_datetime(self, dt=None):
        self.dates.base = to_utc(dt)

    def to_dict(self):
        return dict(
            crfs=[(crf.model, crf.required) for crf in self.crfs],
            crfs_unscheduled=[(crf.model, crf.required) for crf in self.crfs_unscheduled],
            crfs_prn=[(crf.model, crf.required) for crf in self.crfs_prn],
            crfs_missed=[(crf.model, crf.required) for crf in self.crfs_missed],
            requisitions=[
                (requisition.panel.name, requisition.required)
                for requisition in self.requisitions
            ],
            requisitions_unscheduled=[
                (requisition.panel.name, requisition.required)
                for requisition in self.requisitions_unscheduled
            ],
            requisitions_prn=[
                (requisition.model, requisition.required)
                for requisition in self.requisitions_prn
            ],
        )

    def validate_rlower_late(self, utcnow):
        if utcnow - self.rlower_late < utcnow - self.rlower:
            raise VisitDateError(
                "Lower bound error. `Late` lower boundary cannot exceed "
                f"lower boundary. See {self}."
            )

    def validate_rupper_late(self, utcnow):
        if utcnow + self.rupper_late < utcnow + self.rupper:
            raise VisitDateError(
                "Upper bound error. `Late` upper boundary cannot be exceed "
                f"upper boundary. See {self}."
            )



from edc_action_item import ActionWithNotification
from edc_adverse_event.constants import DEATH_REPORT_ACTION
from edc_constants.constants import HIGH_PRIORITY
from edc_ltfu.constants import LTFU_ACTION
from edc_offstudy.constants import END_OF_STUDY_ACTION

from .constants import OFFSCHEDULE_ACTION


class OffscheduleAction(ActionWithNotification):
    name = OFFSCHEDULE_ACTION
    display_name = "Submit Off-Schedule"
    notification_display_name = "Off-Schedule"
    parent_action_names = [
        DEATH_REPORT_ACTION,
        LTFU_ACTION,
    ]
    reference_model = "intecomm_prn.offschedule"
    show_link_to_changelist = True
    admin_site_name = "intecomm_prn_admin"
    priority = HIGH_PRIORITY
    singleton = True

    def get_next_actions(self):
        return [END_OF_STUDY_ACTION]



from dateutil.relativedelta import relativedelta
from django.conf import settings
from django.utils.translation import gettext as _
from edc_utils import convert_php_dateformat, floor_secs, formatted_date, to_utc
from edc_utils.date import to_local

from ..exceptions import (
    ScheduledVisitWindowError,
    ScheduleError,
    UnScheduledVisitWindowError,
)
from .visit_collection import VisitCollectionError

enforce_window_period_enabled = getattr(
    settings, "EDC_VISIT_SCHEDULE_ENFORCE_WINDOW_PERIOD", True
)


class Window:
    def __init__(
        self,
        name=None,
        visits=None,
        dt=None,
        timepoint_datetime=None,
        visit_code=None,
        visit_code_sequence=None,
        baseline_timepoint_datetime=None,
    ):
        self.name = name
        self.visits = visits
        self.timepoint_datetime = to_utc(timepoint_datetime)
        self.dt = to_utc(dt)
        self.visit_code = visit_code
        self.visit_code_sequence = visit_code_sequence
        self.baseline_timepoint_datetime = to_utc(baseline_timepoint_datetime)

    @property
    def datetime_in_window(self):
        if enforce_window_period_enabled:
            if not self.dt:
                raise UnScheduledVisitWindowError("Invalid datetime")
            try:
                self.visits.get(self.visit_code)
            except VisitCollectionError as e:
                raise ScheduleError(e)
            if self.is_scheduled_visit or not self.visits.next(self.visit_code):
                self.raise_for_scheduled_not_in_window()
            else:
                self.raise_for_unscheduled_not_in_window()
        return True

    @property
    def is_scheduled_visit(self):
        return self.visit_code_sequence == 0 or self.visit_code_sequence is None

    def get_window_gap_days(self) -> int:
        days = 0
        visit = self.visits.get(self.visit_code)
        next_visit = self.visits.next(self.visit_code)
        if visit.add_window_gap_to_lower and next_visit:
            days = abs(
                (self.timepoint_datetime + visit.rupper)
                - (self.timepoint_datetime - next_visit.rlower)
            ).days
        return days

    def raise_for_scheduled_not_in_window(self):
        """Returns the datetime if it falls within the
        window period for a scheduled `visit` otherwise
        raises an exception.

        In this case, `visit` is the object from schedule and
        not a model instance.
        """
        visit = self.visits.get(self.visit_code)
        visit.timepoint_datetime = self.timepoint_datetime
        gap_days = self.get_window_gap_days()
        lower = floor_secs(to_utc(visit.dates.lower) - relativedelta(days=gap_days))
        upper = floor_secs(to_utc(visit.dates.upper))
        if not (lower <= floor_secs(to_utc(self.dt)) <= upper):
            lower_date = to_local(visit.dates.lower).strftime(
                convert_php_dateformat(settings.SHORT_DATETIME_FORMAT)
            )
            upper_date = to_local(visit.dates.upper).strftime(
                convert_php_dateformat(settings.SHORT_DATETIME_FORMAT)
            )
            dt = to_local(self.dt).strftime(
                convert_php_dateformat(settings.SHORT_DATETIME_FORMAT)
            )

            raise ScheduledVisitWindowError(
                "Invalid. Date falls outside of the "
                f"window period for this `scheduled` visit. "
                f"Expected a date between {lower_date} "
                f"and {upper_date} for {self.visit_code}. Got `{dt}`."
            )

    def raise_for_unscheduled_not_in_window(self):
        """Returns the datetime if it falls within the
        window period for an unscheduled `visit` otherwise
        raises an exception.

        Window period for an unscheduled date is anytime
        on or after the scheduled date and before the projected
        lower bound of the next visit.

        In this case, `visit` is the object from schedule and
        not a model instance.
        """
        next_visit = self.visits.next(self.visit_code)
        next_timepoint_datetime = self.visits.timepoint_dates(
            dt=self.baseline_timepoint_datetime
        ).get(next_visit)
        if not (
            floor_secs(to_utc(self.dt))
            < floor_secs(to_utc(next_timepoint_datetime - next_visit.rlower))
        ):
            dt_lower = formatted_date(next_timepoint_datetime - next_visit.rlower)
            dt = formatted_date(self.dt)
            raise UnScheduledVisitWindowError(
                _(
                    "Invalid datetime. Falls outside of the "
                    "window period for this `unscheduled` visit. "
                    "Expected a datetime before the next visit. "
                    "Next visit is `%(next_visit_code)s` expected any time "
                    "from `%(dt_lower)s`."
                    "Got `%(visit_code)s`@`%(dt)s`. "
                )
                % dict(
                    next_visit_code=next_visit.code,
                    dt_lower=dt_lower,
                    visit_code=self.visit_code,
                    dt=dt,
                )
            )



from __future__ import annotations

from collections import Counter, defaultdict
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.checks import Error, Warning
from django.core.exceptions import ObjectDoesNotExist

from .site_visit_schedules import site_visit_schedules
from .utils import (
    check_models_in_visit_schedule,
    get_models_from_collection,
    get_proxy_models_from_collection,
    get_proxy_root_model,
)
from .visit import CrfCollection

if TYPE_CHECKING:
    from .visit import Visit


def visit_schedule_check(app_configs, **kwargs):
    errors = []
    if not site_visit_schedules.visit_schedules:
        errors.append(
            Warning("No visit schedules have been registered!", id="edc_visit_schedule.W001")
        )
    site_results = check_models_in_visit_schedule()
    for category, results in site_results.items():
        for result in results:
            if category == "visit_schedules":
                error_code = "E002"
            elif category == "schedules":
                error_code = "E003"
            elif category == "visits":
                error_code = "E004"
            else:
                raise KeyError(f"Unexpected key. Got {category}.")

            errors.append(Warning(result, id=f"edc_visit_schedule.{error_code}"))
    return errors


def check_subject_schedule_history(app_configs, **kwargs) -> list:
    errors = []
    subject_schedule_history_cls = django_apps.get_model(
        "edc_visit_schedule.subjectschedulehistory"
    )
    for obj in subject_schedule_history_cls.objects.all():
        try:
            obj.onschedule_obj
        except LookupError as e:
            errors.append(
                Error(
                    "Invalid onschedule model referenced in SubjectScheduleHistory. "
                    f"See {obj.onschedule_model} for {obj.subject_identifier} "
                    f"Got {e}",
                    id="edc_visit_schedule.E005",
                )
            )
        except ObjectDoesNotExist:
            errors.append(
                "Invalid onschedule model referenced in SubjectScheduleHistory. "
                f"Got {obj.onschedule_model} for {obj.subject_identifier}"
            )
    return errors


def check_onschedule_exists_in_subject_schedule_history(app_configs, **kwargs) -> list:
    errors = []
    subject_schedule_history_cls = django_apps.get_model(
        "edc_visit_schedule.subjectschedulehistory"
    )
    for visit_schedule in site_visit_schedules.visit_schedules.values():
        for schedule in visit_schedule.schedules.values():
            try:
                onschedule_model_cls = getattr(schedule, "onschedule_model_cls")
            except LookupError:
                pass
            else:
                for obj in onschedule_model_cls.objects.all():
                    try:
                        subject_schedule_history_cls.objects.get(
                            subject_identifier=obj.subject_identifier,
                            onschedule_model=onschedule_model_cls._meta.label_lower,
                        )
                    except ObjectDoesNotExist:
                        errors.append(
                            Error(
                                f"Onschedule instance not found in SubjectScheduleHistory. "
                                f"See {obj.subject_identifier} "
                                f"model {onschedule_model_cls._meta.label_lower}."
                            )
                        )
    return errors


def check_form_collections(app_configs, **kwargs):
    errors = []
    for visit_schedule in site_visit_schedules.visit_schedules.values():
        for schedule in visit_schedule.schedules.values():
            for visit in schedule.visits.values():
                for visit_crf_collection, visit_type in [
                    (visit.crfs, "Scheduled"),
                    (visit.crfs_unscheduled, "Unscheduled"),
                    (visit.crfs_missed, "Missed"),
                ]:
                    if proxy_root_alongside_child_err := check_proxy_root_alongside_child(
                        visit=visit,
                        visit_crf_collection=visit_crf_collection,
                        visit_type=visit_type,
                    ):
                        errors.append(proxy_root_alongside_child_err)

                    if same_proxy_root_err := check_multiple_proxies_same_proxy_root(
                        visit=visit,
                        visit_crf_collection=visit_crf_collection,
                        visit_type=visit_type,
                    ):
                        errors.append(same_proxy_root_err)

    return errors


def check_proxy_root_alongside_child(
    visit: Visit,
    visit_crf_collection: CrfCollection,
    visit_type: str,
) -> Error | None:
    all_models = get_models_from_collection(
        collection=visit_crf_collection
    ) + get_models_from_collection(collection=visit.crfs_prn)
    all_proxy_models = get_proxy_models_from_collection(
        collection=visit_crf_collection
    ) + get_proxy_models_from_collection(collection=visit.crfs_prn)

    if child_proxies_alongside_proxy_roots := [
        m for m in all_proxy_models if get_proxy_root_model(m) in all_models
    ]:
        proxy_root_child_pairs = [
            (
                f"proxy_root_model={get_proxy_root_model(proxy)._meta.label_lower}",
                f"proxy_model={proxy._meta.label_lower}",
            )
            for proxy in child_proxies_alongside_proxy_roots
        ]
        return Error(
            "Proxy root model class appears alongside associated child "
            "proxy for a visit. "
            f"Got '{visit}' '{visit_type}' visit Crf collection. "
            f"Proxy root:child models: {proxy_root_child_pairs=}",
            id="edc_visit_schedule.E006",
        )


def check_multiple_proxies_same_proxy_root(
    visit: Visit,
    visit_crf_collection: CrfCollection,
    visit_type: str,
) -> Error | None:
    # Find all proxy models, and map from their 'proxy root' models
    all_proxy_models = get_proxy_models_from_collection(
        collection=visit_crf_collection
    ) + get_proxy_models_from_collection(collection=visit.crfs_prn)
    proxy_root_to_child_proxies: defaultdict[str, list[str]] = defaultdict(list)
    for proxy_model in all_proxy_models:
        child_proxy_model_str = proxy_model._meta.label.lower()
        proxy_root_model_str = get_proxy_root_model(proxy_model)._meta.label.lower()
        proxy_root_to_child_proxies[proxy_root_model_str].append(child_proxy_model_str)

    # Find proxy models declared as sharing a proxy root
    proxies_sharing_roots = get_proxy_models_from_collection(
        collection=CrfCollection(*[f for f in visit_crf_collection if f.shares_proxy_root])
    ) + get_proxy_models_from_collection(
        collection=CrfCollection(*[f for f in visit.crfs_prn if f.shares_proxy_root])
    )
    proxies_sharing_roots_counter = Counter(
        [m._meta.label.lower() for m in proxies_sharing_roots]
    )

    # Filter out valid models/prepare error list
    for proxy_root in list(proxy_root_to_child_proxies):
        proxies_counter = Counter(proxy_root_to_child_proxies[proxy_root])
        if proxies_counter & proxies_sharing_roots_counter == proxies_counter:
            # OK if proxies counter reflects ALL defined proxy shared roots
            del proxy_root_to_child_proxies[proxy_root]
        elif len(proxies_counter) == 1 and next(iter(proxies_counter.values())) <= 2:
            # OK for a single proxy to be defined in two places (CRFs collection + PRNs)
            del proxy_root_to_child_proxies[proxy_root]
        else:
            proxy_root_to_child_proxies[proxy_root].sort()

    if proxy_root_to_child_proxies:
        return Error(
            "Multiple proxies with same proxy root model appear for "
            "a visit. If this is intentional, consider using "
            "`shares_proxy_root` argument when defining Crf. "
            f"Got '{visit}' '{visit_type}' visit Crf collection. "
            f"Proxy root/child models: {dict(proxy_root_to_child_proxies)}",
            id="edc_visit_schedule.E007",
        )



import sys

from django.core.management.color import color_style

style = color_style()


def populate_visit_schedule(sender=None, **kwargs):
    from .models import VisitSchedule
    from .site_visit_schedules import site_visit_schedules

    sys.stdout.write(style.MIGRATE_HEADING("Populating visit schedule:\n"))
    VisitSchedule.objects.update(active=False)
    site_visit_schedules.to_model(VisitSchedule)
    sys.stdout.write("Done.\n")
    sys.stdout.flush()



from django.db import models


class VisitCodeFieldsModelMixin(models.Model):
    visit_code = models.CharField(max_length=25, null=True, editable=False)

    visit_code_sequence = models.IntegerField(
        verbose_name="Sequence",
        default=0,
        null=True,
        blank=True,
        help_text=(
            "An integer to represent the sequence of additional "
            "appointments relative to the base appointment, 0, needed "
            "to complete data collection for the timepoint. (NNNN.0)"
        ),
    )

    class Meta:
        abstract = True



from .visit_code_fields_model_mixin import VisitCodeFieldsModelMixin
from .visit_schedule_fields_model_mixin import VisitScheduleFieldsModelMixin
from .visit_schedule_methods_model_mixin import (
    VisitScheduleMethodsModelMixin,
    VisitScheduleModelMixinError,
)
from .visit_schedule_model_mixins import VisitScheduleModelMixin



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .exceptions import SiteVisitScheduleError, VisitScheduleBaselineError
from .site_visit_schedules import site_visit_schedules

if TYPE_CHECKING:
    from decimal import Decimal

    from .schedule import Schedule
    from .visit_schedule import VisitSchedule


__all__ = ["Baseline"]


class Baseline:
    def __init__(
        self,
        instance: Any = None,
        timepoint: Decimal | None = None,
        visit_code_sequence: int | None = None,
        visit_schedule_name: str | None = None,
        schedule_name: str | None = None,
    ):
        if instance:
            try:
                instance = instance.appointment
            except AttributeError:
                try:
                    instance = instance.subject_visit.appointment
                except AttributeError:
                    pass
            self.visit_schedule_name = instance.visit_schedule_name
            self.schedule_name = instance.schedule_name
            self.visit_code_sequence = instance.visit_code_sequence
            self.timepoint = instance.timepoint
        else:
            self.visit_schedule_name = visit_schedule_name
            self.schedule_name = schedule_name
            self.visit_code_sequence = visit_code_sequence
            self.timepoint = timepoint
            if self.timepoint is None:
                raise VisitScheduleBaselineError("timepoint may not be None")
        if not any([x == self.timepoint for x in self.timepoints.values()]):
            raise VisitScheduleBaselineError(
                f"Unknown timepoint. For schedule {self.visit_schedule}.{self.schedule}. "
                f"Got {self.timepoint} not in {self.timepoints}"
            )
        self.value: bool = (
            self.timepoint == self.baseline_timepoint and self.visit_code_sequence == 0
        )

    @property
    def visit_schedule(self) -> VisitSchedule:
        self.have_required_attrs_or_raise()
        try:
            visit_schedule = site_visit_schedules.get_visit_schedule(self.visit_schedule_name)
        except SiteVisitScheduleError as e:
            raise VisitScheduleBaselineError(str(e))
        return visit_schedule

    @property
    def schedule(self) -> Schedule:
        try:
            schedule = self.visit_schedule.schedules.get(self.schedule_name)
        except SiteVisitScheduleError as e:
            raise VisitScheduleBaselineError(str(e))
        return schedule

    @property
    def baseline_timepoint(self) -> Decimal:
        """Returns a decimal that is the first timepoint in this schedule"""
        return self.schedule.visits.first.timepoint

    @property
    def timepoints(self) -> dict:
        return self.schedule.visits.timepoints

    def have_required_attrs_or_raise(self):
        data = {
            k: getattr(self, k, None) is None
            for k in [
                "visit_schedule_name",
                "schedule_name",
                "visit_code_sequence",
                "timepoint",
            ]
        }

        if any(data.values()):
            raise VisitScheduleBaselineError(
                "Missing value(s). Unable to determine if baseline. "
                f"Got `None` for {[k for k, v in data.items() if v is True]}."
            )



from edc_sites.single_site import SingleSite
from edc_sites.site import sites

suffix = "clinicedc.org"
language_codes = ["en"]
sites.register(
    SingleSite(
        10,
        "mochudi",
        title="Mochudi",
        country="botswana",
        country_code="bw",
        language_codes=language_codes,
        domain=f"mochudi.bw.{suffix}",
    ),
    SingleSite(
        20,
        "molepolole",
        title="Molepolole",
        country="botswana",
        country_code="bw",
        language_codes=language_codes,
        domain=f"molepolole.bw.{suffix}",
    ),
    SingleSite(
        30,
        "lobatse",
        title="Lobatse",
        country="botswana",
        country_code="bw",
        language_codes=language_codes,
        domain=f"lobatse.bw.{suffix}",
    ),
    SingleSite(
        40,
        "gaborone",
        title="Gaborone",
        country="botswana",
        country_code="bw",
        language_codes=language_codes,
        domain=f"gaborone.bw.{suffix}",
    ),
    SingleSite(
        50,
        "karakobis",
        title="Karakobis",
        country="botswana",
        country_code="bw",
        language_codes=language_codes,
        domain=f"karakobis.bw.{suffix}",
    ),
    SingleSite(
        60,
        "windhoek",
        title="Windhoek",
        country="namibia",
        country_code="na",
        language_codes=language_codes,
        domain=f"windhoek.bw.{suffix}",
    ),
)



from django.contrib import admin

from .admin_site import visit_schedule_app_admin
from .models import CrfOne


@admin.register(CrfOne, site=visit_schedule_app_admin)
class CrfOneAdmin(admin.ModelAdmin):
    pass



from django.urls.conf import path, re_path

from .admin_site import edc_visit_schedule_admin
from .views import HomeView, VisitScheduleView

app_name = "edc_visit_schedule"

urlpatterns = [
    path("admin/", edc_visit_schedule_admin.urls),
    re_path(
        r"visit_schedule/(?P<visit_schedule>[0-9A-Za-z_]+)/"
        "(?P<schedule>^[0-9A-Za-z_]+$)/(?P<visit_code>^[0-9]+$)/",
        VisitScheduleView.as_view(),
        name="visit_schedule_url",
    ),
    re_path(
        r"visit_schedule/(?P<visit_schedule>[0-9A-Za-z_]+)/(?P<schedule>^[0-9A-Za-z_]+$)/",
        VisitScheduleView.as_view(),
        name="visit_schedule_url",
    ),
    re_path(
        r"visit_schedule/(?P<visit_schedule>[0-9A-Za-z_]+)/",
        VisitScheduleView.as_view(),
        name="visit_schedule_url",
    ),
    re_path(
        r"visit_schedule/",
        VisitScheduleView.as_view(),
        name="visit_schedule_url",
    ),
    path(r"", HomeView.as_view(), name="home_url"),
]



from .crf import Crf
from .crf_collection import CrfCollection
from .forms_collection import FormsCollection, FormsCollectionError
from .requisition import Requisition
from .requisition_collection import RequisitionCollection
from .visit import Visit, VisitCodeError, VisitDateError
from .window_period import WindowPeriod



from django.apps import apps as django_apps


class InvalidModel(Exception):
    pass


class SimpleModelValidator:
    """Validate label lower without loading the model.

    Does not validate the model_name other than it cannot be None.
    """

    def __init__(self, model=None, attr=None):
        try:
            app_label, _ = model.split(".")
        except AttributeError:
            raise InvalidModel(f"Invalid label lower format for '{attr}'. " f"Got {model}")
        else:
            app_labels = [app_config.name for app_config in django_apps.get_app_configs()]
            if app_label not in app_labels:
                raise InvalidModel(
                    f"Invalid model. app_label does not exist for " f"'{attr}'. Got {model}"
                )



from django.test import TestCase

from edc_visit_schedule.visit import Crf
from edc_visit_schedule.visit.crf import CrfModelNotProxyModelError


class TestCrfCollection(TestCase):
    def test_crf_ok(self):
        try:
            Crf(show_order=1, model="visit_schedule_app.CrfOne")
        except Exception as e:
            self.fail(f"Exception unexpectedly raised. Got {e}")

        try:
            Crf(show_order=1, model="visit_schedule_app.CrfOneProxyOne")
        except Exception as e:
            self.fail(f"Exception unexpectedly raised. Got {e}")

        try:
            Crf(show_order=1, model="visit_schedule_app.CrfTwo")
        except Exception as e:
            self.fail(f"Exception unexpectedly raised. Got {e}")

        try:
            Crf(show_order=1, model="visit_schedule_app.CrfThree")
        except Exception as e:
            self.fail(f"Exception unexpectedly raised. Got {e}")

    def test_proxy_child_crf_with_allow_proxy_parent_clash_ok(self):
        try:
            Crf(
                show_order=1,
                model="visit_schedule_app.CrfOneProxyOne",
                shares_proxy_root=True,
            )
        except Exception as e:
            self.fail(f"Exception unexpectedly raised. Got {e}")

    def test_proxy_root_crf_with_allow_proxy_parent_clash_raises(self):
        with self.assertRaises(CrfModelNotProxyModelError) as cm:
            Crf(show_order=1, model="visit_schedule_app.CrfOne", shares_proxy_root=True)
        self.assertIn(
            "Invalid use of `shares_proxy_root=True`. CRF model is not a proxy model.",
            str(cm.exception),
        )
        self.assertIn("visit_schedule_app.crfone", str(cm.exception))

    def test_non_proxy_crf_with_allow_proxy_parent_clash_raises(self):
        with self.assertRaises(CrfModelNotProxyModelError) as cm:
            Crf(
                show_order=1,
                model="visit_schedule_app.CrfThree",
                shares_proxy_root=True,
            )
        self.assertIn(
            "Invalid use of `shares_proxy_root=True`. CRF model is not a proxy model.",
            str(cm.exception),
        )
        self.assertIn("visit_schedule_app.crfthree", str(cm.exception))
